(function ($) {
    $.jelastic = $.jelastic || {};
    $.jelastic.polling = $.jelastic.polling || {};

    $.jelastic.polling.ActionPoller = function (options) {
        var $eventBus = $({}),
            actionKey = $.jelastic.utils.ActionKey(),
            actionHandler,
            trackingService,
            pollTimeoutId,
            sleepTimeoutId,
            self,
            state;

        options = $.extend({
            baseUrl: null,
            session: null,
            activeInterval: 1000,
            sleepingInterval: 10000,
            retriesCount: 3,
            reloadCount: 4,
            onActionStarted: null,
            onActionCompleted: null
        }, options);

        state = {
            firstStart: true,
            enabled: false,
            active: false,
            startTime: null,
            pollsCount: 0,
            pollsBeforeReload: 0,
            actionKeys: {},
            actions: $.jelastic.data.MixedCollection(),
            buffer: $.jelastic.data.MixedCollection(),
        };

        trackingService = $.jelastic.api.TrackingService({
            baseUrl: options.baseUrl,
            session: options.session
        });

        self = {
            options: options,

            start: function () {
                state.enabled = true;

                self.resetTimers();
                self.loadActionsHistory()
                    .done($.proxy(self.poll, self));

                return self;
            },

            stop: function () {
                state.enabled = false;

                return self.resetTimers();
            },

            reset: function () {
                state.pollsCount = 0;
                state.pollsBeforeReload = 0;

                if (!state.enabled) {
                    self.start();
                }
            },

            addCompositeActionHandler: function (name, handler) {
                if (!actionHandler) {
                    actionHandler = $.jelastic.polling.ActionHandler(self);
                }

                actionHandler.add(name, handler);
            },

            /**
             * @private
             * @return {ActionPoller}
             */
            poll: function () {
                // console.log("next poll");

                trackingService.getCurrentActions().done(function (response) {
                    var currentActions;

                    if (response.result === 0) {
                        currentActions = $.jelastic.data.MixedCollection(response.array);

                        self.checkCompletedActions(currentActions)
                            .checkNewActions(currentActions)
                            .scheduleNextPoll(currentActions);
                    } else {
                        self.scheduleNextPoll();
                    }
                });

                return self;
            },

            /**
             * @private
             * @param {MixedCollection} [currentActions]
             */
            scheduleNextPoll: function (currentActions) {
                state.pollsBeforeReload += 1;

                if (state.pollsBeforeReload > options.reloadCount) {
                    state.pollsBeforeReload = 0;
                }

                if (currentActions && currentActions.length > 0) {
                    state.pollsCount = 0;

                    if (state.pollsBeforeReload === options.reloadCount) {
                        self.loadActionsHistory();
                    }

                    pollTimeoutId =  setTimeout(self.poll, options.activeInterval);
                } else {
                    if (state.pollsCount < options.retriesCount) {
                        state.pollsCount += 1;
                        pollTimeoutId = setTimeout(self.poll, options.activeInterval);
                    } else {
                        self.loadActionsHistory();
                        sleepTimeoutId = setTimeout(self.poll, options.sleepingInterval);
                    }
                }
            },

            /**
             * @private
             * @return {ActionPoller}
             */
            resetTimers: function () {
                if (pollTimeoutId) {
                    clearTimeout(pollTimeoutId);
                    pollTimeoutId = null;
                }

                if (sleepTimeoutId) {
                    clearTimeout(sleepTimeoutId);
                    sleepTimeoutId = null;
                }

                return self;
            },

            /**
             * @private
             * @param {MixedCollection} currentActions
             * @return {ActionPoller}
             */
            checkCompletedActions: function (currentActions) {
                var actionBuffer = state.buffer,
                    completedActions,
                    actionId,
                    action,
                    count,
                    i;

                if (actionBuffer.length > 0) {
                    completedActions = [];

                    for (i = 0; i < actionBuffer.length; i += 1) {
                        actionId = actionBuffer.getKeyAt(i);

                        if (!currentActions.containsKey(actionId)) {
                            completedActions.push(actionId);
                        }
                    }

                    count = completedActions.length;

                    if (count > 0) {
                        for (i = 0; i < count; i += 1) {
                            action = actionBuffer.getItem(actionId);
                            actionBuffer.remove(action);
                        }

                        if (actionBuffer.length < count) {
                            self.loadActionsHistory();
                        }
                    }
                }

                return self;
            },

            /**
             * @private
             * @param {MixedCollection} currentActions
             * @return {ActionPoller}
             */
            checkNewActions: function (currentActions) {
                var actionStore = state.actions,
                    actionBuffer = state.buffer,
                    action,
                    i;

                for (i = 0; i < currentActions.length; i += 1) {
                    action = currentActions.getAt(i);

                    if (!actionStore.containsKey(action.id)) {
                        actionStore.insert(0, action);
                        actionBuffer.add(action);

                        self.onActionStarted(action);
                    }
                }

                return self;
            },

            /**
             * @private
             * @param {Date} date
             */
            formatDate: (function() {
                function normalize(datePart) {
                    return (datePart + "").length === 1 ? "0" + datePart : datePart;
                }

                return function (date, seconds) {
                    return date.getUTCFullYear() + "-" +
                        normalize(date.getUTCMonth() + 1) + "-" +
                        normalize(date.getUTCDate()) + " " +
                        normalize(date.getUTCHours()) + ":" +
                        normalize(date.getUTCMinutes()) + ":" +
                        (seconds !== false ? normalize(date.getUTCSeconds()) : "00");
                };
            }()),

            /**
             * @private
             * @return {Promise}
             */
            loadActionsHistory: function () {
                var count = state.startTime ? null : 10,
                    deferred = $.Deferred();

                // console.log("load history");

                trackingService.getUidActions(state.startTime, null, count).done(function (response) {
                    var actions,
                        receivedAction,
                        storedAction,
                        actionId,                        
                        i;

                    if (response.result === 0) {
                        actions = $.jelastic.data.MixedCollection(response.array);

                        if (actions.length > 0) {
                            for (i = 0; i < actions.length; i += 1) {
                                receivedAction = actions.getAt(i);
                                actionId = receivedAction.id;
                                storedAction = state.actions.getItem(actionId);

                                if (storedAction) {
                                    if (typeof storedAction.result === "undefined") {
                                        self.onActionCompleted(receivedAction);

                                        state.actions.replace(actionId, receivedAction);
                                        state.buffer.removeKey(actionId);
                                    }
                                } else {
                                    state.actions.insert(0, receivedAction);
                                    
                                    if (!state.firstStart) {
                                        self.onActionStarted(receivedAction);
                                        self.onActionCompleted(receivedAction);
                                    }
                                }
                            }

                            state.firstStart = false;
                            state.startTime = self.formatDate(
                                new Date(actions[state.startTime ? "last" : "first"]().starttime),
                                false
                            );
                        }
                    }

                    deferred.resolve(response);
                });

                return deferred.promise();
            },

            resolveCompositeAction: function (trackedAction, isCompleted) {
                var FINISH_ACTION_ID = 30,
                    data;

                if (trackedAction.actionkey) {
                    data = actionKey.decode(trackedAction.actionkey);
                    trackedAction.actionKeyPayload = data;

                    if (!state.actionKeys[data.id]) {
                        $eventBus.trigger("composite-action-started", [trackedAction.actionKeyPayload, trackedAction]);
                    }

                    if (isCompleted) {
                        if (trackedAction.action === FINISH_ACTION_ID) {
                            $eventBus.trigger("internal-action-completed", [trackedAction.actionKeyPayload, trackedAction]);
                        } else if (trackedAction.result !== 0) {
                            $eventBus.trigger("internal-action-error", [trackedAction.actionKeyPayload, trackedAction]);
                        }
                    }

                    state.actionKeys[data.id] = true;
                }
            },

            onActionStarted: function (action) {
                self.resolveCompositeAction(action);

                $eventBus.trigger("action-started", [action, action.text]);
            },

            onActionCompleted: function (action) {
                if (!state.firstStart) {
                    self.resolveCompositeAction(action, true);

                    $eventBus.trigger("action-completed", [action, action.text]);
                }
            },

            on: function (name, callback, scope) {
                if ($.type(callback) === "function") {
                    $eventBus.on(name, function () {
                        var args = Array.prototype.slice.call(arguments);

                        args.shift();
                        callback.apply(scope || window, args);
                    });
                }
            }
        };

        return self;
    };
})(jQuery);